# Lint as: python3
"""The base class for all quadrupeds."""
from typing import Any, Callable, Dict, Sequence, Tuple, Text, Union
import gin
import gym
import numpy as np

from pybullet_utils import bullet_client
from pybullet_envs.minitaur.envs_v2.sensors import sensor as sensor_lib
from pybullet_envs.minitaur.robots import hybrid_motor_model
from pybullet_envs.minitaur.robots import robot_base
from pybullet_envs.minitaur.robots import robot_config
from pybullet_envs.minitaur.robots import robot_urdf_loader
from pybullet_envs.minitaur.robots.safety import data_types as safety_data_types
from pybullet_envs.minitaur.robots.utilities import kinematics_utils

_UNIT_QUATERNION = (0, 0, 0, 1)
_GRAVITY_ACCELERATION_OFFSET = (0, 0, 10)


@gin.configurable
class QuadrupedBase(robot_base.RobotBase):
  """The basic quadruped class for both sim and real robots."""

  def __init__(
      self,
      pybullet_client: bullet_client.BulletClient,
      clock: Callable[..., float],
      motor_control_mode: robot_config.MotorControlMode,
      motor_limits: robot_config.MotorLimits,
      motor_model_class: Any = hybrid_motor_model.HybridMotorModel,
      action_filter: Any = None,
      sensors: Sequence[sensor_lib.Sensor] = (),
      safety_config: safety_data_types.SafetyConfig = None,
      **kwargs,
  ):
    """Initializes the class.

    Args:
      pybullet_client: The PyBullet client.
      clock: The sim or real clock. The clock function is typically provided by
        the gym environment.
      motor_control_mode: Specifies in which mode the motor operates.
      motor_limits: The motor limits of the robot. Used by the motor_model_class
        and action space building.
      motor_model_class: The motor model to use. Not needed for real robots.
      action_filter: The filter to smooth and/or regulate the actions.
      sensors: All sensors mounted on the robot.
      safety_config: The safety setting for the robot.
      **kwargs: Additional args.
    """

    self._pybullet_client = pybullet_client
    self._clock = clock
    self._motor_control_mode = motor_control_mode
    self._motor_model_class = motor_model_class
    self._motor_limits = motor_limits
    self._action_space = None
    self._action_names = None
    self._action_filter = action_filter
    self._sensors = sensors
    self._safety_config = safety_config
    self._urdf_loader = None
    self._last_base_velocity = np.zeros(3)
    self._last_observation_time = self._clock()
    self._last_base_acceleration_world = np.zeros(3)
    self._last_base_acceleration_accelerometer = np.zeros(3)

    self.load()

  def load(
      self,
      base_position: Tuple[float] = None,
      base_orientation_quaternion: Tuple[float] = None,
      joint_angles: Union[Dict[Text, float], Tuple[float]] = None,
  ):
    """Loads the URDF with the configured pose.

    Args:
      base_position: The base position after URDF loading. Will use the
        configured pose in gin if None.
      base_orientation_quaternion: The base orientation after URDF loading. Will
        use the configured values in gin if not specified.
      joint_angles: The desired joint angles after loading. Will use the
        configured values if None.
    """
    # A robot specific pre loading routing.
    self._pre_load()

    if not self._urdf_loader:
      self._urdf_loader = robot_urdf_loader.RobotUrdfLoader(
          pybullet_client=self._pybullet_client)

    # Record the urdf pose at loading, which will be used as the rotation
    # reference for base rotation computation.
    self._init_urdf_position, self._init_orientation_quat = (
        self._pybullet_client.getBasePositionAndOrientation(
            self._urdf_loader.robot_id))
    unused_position, self._init_orientation_inv_quat = (
        self._pybullet_client.invertTransform(
            position=(0, 0, 0), orientation=self._init_orientation_quat))

    # Joint ids may be different from the motor ids.
    self._joint_id_dict = self._urdf_loader.get_joint_id_dict()
    for joint_id in self._joint_id_dict.values():
      # Disables the default motors in PyBullet.
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self._urdf_loader.robot_id,
          jointIndex=joint_id,
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=0)
      # Removes the default joint damping in PyBullet.
      self._pybullet_client.changeDynamics(
          self._urdf_loader.robot_id,
          joint_id,
          linearDamping=0,
          angularDamping=0)

    # We expect that this is non-empty for all quadrupedes, and should be an
    # OrderedDict.
    self._motor_id_dict = self._urdf_loader.get_motor_id_dict()
    if not self._motor_id_dict:
      raise ValueError("Motor id dict cannot be empty for quadrupeds.")
    self._motor_ids = self._motor_id_dict.values()
    self._num_motors = len(self._motor_id_dict)

    self._build_action_space()

    # Not needed for real robots.
    if self._motor_model_class:
      # TODO(b/151664871): Also supports position/velocity limits in the motor
      # model.
      self._motor_model = self._motor_model_class(
          num_motors=self._num_motors,
          motor_control_mode=self._motor_control_mode,
          torque_lower_limits=self._motor_limits.torque_lower_limits,
          torque_upper_limits=self._motor_limits.torque_upper_limits,
      )

    # Caches the variable for faster computation during stepping.
    self._motor_direction_dict = self._urdf_loader.get_joint_direction_dict(
        self._motor_id_dict.keys())
    self._motor_directions = np.array(list(self._motor_direction_dict.values()))

    self._motor_offset_dict = self._urdf_loader.get_joint_offset_dict(
        self._motor_id_dict.keys())
    self._motor_offsets = np.array(list(self._motor_offset_dict.values()))

    # A robot specific routine post loading.
    self._on_load()

    # Robot sensors may use information from the class. So we initialize them
    # after the loading is done.
    for sensor in self._sensors:
      sensor.set_robot(self)

  def _build_action_space(self):
    """Builds the action space of the robot using the motor limits."""
    if self._motor_control_mode == robot_config.MotorControlMode.POSITION:
      self._action_space = gym.spaces.Box(
          low=self._motor_limits.angle_lower_limits,
          high=self._motor_limits.angle_upper_limits,
          shape=(self._num_motors,),
          dtype=np.float32)  # TODO(b/159160184) Make dtype configurable.
      self._action_names = tuple(
          "POSITION_{}".format(motor) for motor in self._motor_id_dict.keys())
    elif self._motor_control_mode == robot_config.MotorControlMode.TORQUE:
      self._action_space = gym.spaces.Box(
          low=self._motor_limits.torque_lower_limits,
          high=self._motor_limits.torque_upper_limits,
          shape=(self._num_motors,),
          dtype=np.float32)
      self._action_names = tuple(
          "TORQUE_{}".format(motor) for motor in self._motor_id_dict.keys())
    elif self._motor_control_mode == robot_config.MotorControlMode.HYBRID:
      hybrid_action_limits_low = [
          self._motor_limits.angle_lower_limits,  # q
          # q_dot
          self._motor_limits.velocity_lower_limits,
          0,  # kp
          0,  # kd
          self._motor_limits.torque_lower_limits
      ]  # tau
      hybrid_action_limits_high = [
          self._motor_limits.angle_upper_limits,
          self._motor_limits.velocity_upper_limits, np.inf, np.inf,
          self._motor_limits.torque_upper_limits
      ]
      space_low = np.full(
          (self._num_motors, robot_config.HYBRID_ACTION_DIMENSION),
          hybrid_action_limits_low).ravel()
      space_high = np.full(
          (self._num_motors, robot_config.HYBRID_ACTION_DIMENSION),
          hybrid_action_limits_high).ravel()
      self._action_space = gym.spaces.Box(
          low=space_low, high=space_high, dtype=np.float32)
      self._action_names = tuple(
          "HYBRID_{}".format(motor) for motor in self._motor_id_dict.keys())
    else:
      raise NotImplementedError("Not yet implemented!")

  def _pre_load(self):
    """Robot specific pre load routine.

    For example, this allows configuration of the URDF loader.
    """
    pass

  def _on_load(self):
    """Robot specific post load routine.

    For example, we need to add add additional hinge constraints to the leg
    components of Minitaur after loading.

    """
    pass

  @gin.configurable
  def reset(
      self,
      base_position: Tuple[float] = None,
      base_orientation_quaternion: Tuple[float] = None,
      joint_angles: Union[Dict[Text, float], Tuple[float]] = None,
      save_base_pose: bool = False,
      **kwargs,
  ):
    """Resets the robot base and joint pose without reloading the URDF.

    Base pose resetting only works for simulated robots or visualization of real
    robots. This routine also updates the initial observation dict.

    Args:
      base_position: The desired base position. Will use the configured pose in
        gin if None. Does not affect the position of the real robots in general.
      base_orientation_quaternion: The base orientation after resetting. Will
        use the configured values in gin if not specified.
      joint_angles: The desired joint angles after resetting. Will use the
        configured values if None.
      save_base_pose: Save the base position and orientation as the default pose
        after resetting.
      **kwargs: Other args for backward compatibility. TODO(b/151975607): Remove
        after migration.
    """
    # Reset the robot's motor model.
    self._motor_model.reset()

    # Reset the quantities for computing base acceleration.
    self._last_base_velocity = np.zeros(3)
    self._last_observation_time = self._clock()
    self._last_base_acceleration_world = np.zeros(3)
    self._last_base_acceleration_accelerometer = np.zeros(3)

    # Solves chicken and egg problem. We need to run a control step to obtain
    # the first motor torques.
    self._motor_torques = np.zeros(self._num_motors)

    # Receives a set of observation from the robot in case the reset function
    # needs to use them.
    self.receive_observation()

    self._reset_base_pose(base_position, base_orientation_quaternion)
    self._reset_joint_angles(joint_angles)

    if save_base_pose:
      # Records the base pose at resetting again, in case Reset is called with a
      # different base orientation. This base pose will be used as zero
      # rotation reference for base rotation computation.
      self._init_urdf_position, self._init_orientation_quat = (
          self._pybullet_client.getBasePositionAndOrientation(
              self._urdf_loader.robot_id))
      unused_position, self._init_orientation_inv_quat = (
          self._pybullet_client.invertTransform(
              position=(0, 0, 0), orientation=self._init_orientation_quat))

    # Updates the observation at the end of resetting.
    self.receive_observation()
    self._time_at_reset = self._clock()

  def GetTimeSinceReset(self):
    return self._clock() - self._time_at_reset

  def _reset_base_pose(self, position=None, orientation_quat=None):
    """Resets the pose of the robot's base.

    Base pose resetting only works for simulated robots or visualization of real
    robots.

    Args:
      position: The desired base position. Will use the configured pose in gin
        if None.
      orientation_quat: The desired base rotation. Will use the configured
        default pose in None.
    """
    self._urdf_loader.reset_base_pose(position, orientation_quat)

  def _reset_joint_angles(self,
                          joint_angles: Union[Tuple[float],
                                              Dict[Text, float]] = None):
    """Resets the joint pose.

    Real robots need to specify their routine to send joint angles. Simulated
    Minitaur robots also needs to use dynamics to drive the motor joints, due to
    the additional hinge joints not present in the URDF.

    Args:
      joint_angles: The joint pose if provided. Will use the robot default pose
        from configuration.
    """
    # TODO(b/148897311): Supports tuple as the input.
    self._urdf_loader.reset_joint_angles(joint_angles)

  def terminate(self):
    """The safe exit routine for the robot.

    Only implemented for real robots.

    """
    pass

  def step(self, action: Any, num_sub_steps: int = 1):
    """Steps the simulation.

    This is maintained for backward compatibility with the old robot class.

    Args:
      action: The control command to be executed by the robot.
      num_sub_steps: Each action can be applied (possibly with interpolation)
        multiple timesteps, to simulate the elapsed time between two consecutive
        commands on real robots.
    """
    action = self.pre_control_step(action)

    for _ in range(num_sub_steps):
      # TODO(b/149252003): Add sub sampling.
      self.apply_action(action)
      # Timestep is pre-determined at simulation setup.
      self._pybullet_client.stepSimulation()
      self.receive_observation()

    self.post_control_step()

  def pre_control_step(self, action: Any, control_timestep: float = None):
    """Processes the action and updates per control step quantities.

    Args:
      action: The input control command.
      control_timestep: The control time step in the environment.
        TODO(b/153835005), we can remove this once we pass env to the robot.

    Returns:
      The filtered action.
    """
    if self._action_filter:
      # We assume the filter will create a set of interpolated results.
      action = self._action_filter.filter(action)
    return action

  def apply_action(self, motor_commands, motor_control_mode=None):

    # TODO(b/148897311): Supports dict in the future.
    motor_commands = np.asarray(motor_commands)

    # We always use torque based control at the lowest level for quadrupeds.
    unused_observed_torques, actual_torques = (
        self._motor_model.get_motor_torques(motor_commands, motor_control_mode))
    self._motor_torques = actual_torques

    # Converts the motor torques to URDF joint space, which may have different
    # directions.
    applied_motor_torques = np.multiply(actual_torques, self._motor_directions)

    self._pybullet_client.setJointMotorControlArray(
        bodyIndex=self._urdf_loader.robot_id,
        jointIndices=self._motor_ids,
        controlMode=self._pybullet_client.TORQUE_CONTROL,
        forces=applied_motor_torques)

  def _get_base_roll_pitch_yaw_rate(self):
    _, angular_velocity = self._pybullet_client.getBaseVelocity(
        self._urdf_loader.robot_id)
    return kinematics_utils.rotate_to_base_frame(
        self._pybullet_client, self.urdf_loader.robot_id, angular_velocity,
        self._init_orientation_inv_quat)

  def _get_base_velocity(self):
    base_velocity, _ = self._pybullet_client.getBaseVelocity(
        self._urdf_loader.robot_id)
    return base_velocity

  def _update_base_acceleration(self):
    """Update the base acceleration using finite difference."""
    if self._last_observation_time < self.timestamp:
      self._last_base_acceleration_world = (
          np.array(self._base_velocity) - self._last_base_velocity) / (
              self.timestamp - self._last_observation_time)
      _, inv_base_orientation = self.pybullet_client.invertTransform(
          np.zeros(3), np.array(self.base_orientation_quaternion))

      # An offset is added to the acceleration measured in the world frame
      # because the accelerometer reading is in the frame of free-falling robot.
      base_acceleration_accelerometer = self.pybullet_client.multiplyTransforms(
          np.zeros(3), inv_base_orientation,
          self._last_base_acceleration_world + _GRAVITY_ACCELERATION_OFFSET,
          _UNIT_QUATERNION)[0]
      self._last_base_acceleration_accelerometer = np.array(
          base_acceleration_accelerometer)

  def receive_observation(self):
    """Receives the observations for all sensors."""
    # Update the intrinsic values including the joint angles, joint
    # velocities, and imu readings.
    self._base_position, base_orientation_quat = (
        self._pybullet_client.getBasePositionAndOrientation(
            self._urdf_loader.robot_id))
    _, self._base_orientation_quat = self._pybullet_client.multiplyTransforms(
        positionA=(0, 0, 0),
        orientationA=self._init_orientation_inv_quat,
        positionB=(0, 0, 0),
        orientationB=base_orientation_quat)
    self._base_velocity = self._get_base_velocity()
    self._base_roll_pitch_yaw = self._pybullet_client.getEulerFromQuaternion(
        self._base_orientation_quat)

    self._base_roll_pitch_yaw_rate = self._get_base_roll_pitch_yaw_rate()

    self._joint_states = self._pybullet_client.getJointStates(
        self._urdf_loader.robot_id, self._motor_ids)
    self._motor_angles = np.array(
        [joint_state[0] for joint_state in self._joint_states])
    self._motor_angles = (self._motor_angles -
                          self._motor_offsets) * self._motor_directions

    self._motor_velocities = np.array(
        [joint_state[1] for joint_state in self._joint_states])
    self._motor_velocities = self._motor_velocities * self._motor_directions

    # We use motor models to track the delayed motor positions and velocities
    # buffer.
    if self._motor_model:
      self._motor_model.update(self._clock(), self._motor_angles,
                               self._motor_velocities)

    self._update_base_acceleration()
    # Update the latest base velocity and timestamp at the end of the API.
    self._last_base_velocity = np.array(self._base_velocity)
    self._last_observation_time = self.timestamp

  def post_control_step(self):
    """Called at the end of a control step outside the action repeat loop."""
    pass

  # TODO(tingnan): Change from "foot_positions" to "feet_positions".
  def motor_angles_from_foot_positions(self,
                                       foot_positions,
                                       position_in_world_frame=False):
    """Use IK to compute the motor angles, given the feet links' positions.

    Args:
      foot_positions: The foot links' positions in frame specified by the next
        parameter. The input is a numpy array of size (4, 3).
      position_in_world_frame: Whether the foot_positions are specified in the
        world frame.

    Returns:
      A tuple. The position indices and the angles for all joints along the
      leg. The position indices is consistent with the joint orders as returned
      by GetMotorAngles API.
    """
    joint_position_idxs = np.arange(self.num_motors)
    foot_link_ids = tuple(self._urdf_loader.get_end_effector_id_dict().values())
    joint_angles = kinematics_utils.joint_angles_from_link_positions(
        pybullet_client=self.pybullet_client,
        urdf_id=self.robot_id,
        link_positions=foot_positions,
        link_ids=foot_link_ids,
        joint_dof_ids=joint_position_idxs,
        positions_are_in_world_frame=position_in_world_frame)
    joint_angles = np.multiply(
        np.asarray(joint_angles) - np.asarray(self._motor_offsets),
        self._motor_directions)
    return joint_position_idxs, joint_angles

  # TODO(tingnan): Change from "foot_positions" to "feet_positions".
  def foot_positions(self, position_in_world_frame=False):
    """Returns the robot's foot positions in the base/world frame."""
    foot_positions = []
    foot_link_ids = tuple(self._urdf_loader.get_end_effector_id_dict().values())
    for foot_id in foot_link_ids:
      if not position_in_world_frame:
        foot_positions.append(
            kinematics_utils.link_position_in_base_frame(
                pybullet_client=self.pybullet_client,
                urdf_id=self.robot_id,
                link_id=foot_id,
            ))
      else:
        foot_positions.append(
            kinematics_utils.link_position_in_world_frame(
                pybullet_client=self.pybullet_client,
                urdf_id=self.robot_id,
                link_id=foot_id,
            ))
    return np.array(foot_positions)

  def feet_contact_forces(self) -> Sequence[np.ndarray]:
    """Gets the contact forces on all feet.

    Reals robot may use a robot specific implementation. For example, the
    Laikago will measure each contact force in the corresponding foot's local
    frame, and this force will not be the total contact force due to the sensor
    limitation.

    For simulated robots, we wll always report the force in the base frame.

    Returns:
      A list of foot contact forces.
    """
    foot_link_ids = tuple(self._urdf_loader.get_end_effector_id_dict().values())
    contact_forces = [np.zeros(3) for _ in range(len(foot_link_ids))]

    all_contacts = self._pybullet_client.getContactPoints(
        bodyA=self._urdf_loader.robot_id)

    for contact in all_contacts:
      (unused_flag, body_a_id, body_b_id, link_a_id, unused_link_b_id,
       unused_pos_on_a, unused_pos_on_b, contact_normal_b_to_a, unused_distance,
       normal_force, friction_1, friction_direction_1, friction_2,
       friction_direction_2) = contact

      # Ignore self contacts
      if body_b_id == body_a_id:
        continue

      if link_a_id in foot_link_ids:
        normal_force = np.array(contact_normal_b_to_a) * normal_force
        friction_force = np.array(friction_direction_1) * friction_1 + np.array(
            friction_direction_2) * friction_2
        force = normal_force + friction_force
        local_force = kinematics_utils.rotate_to_base_frame(
            self._pybullet_client, self.urdf_loader.robot_id, force,
            self._init_orientation_inv_quat)
        local_force_norm = np.linalg.norm(local_force)
        toe_link_order = foot_link_ids.index(link_a_id)
        if local_force_norm > 0:
          contact_forces[toe_link_order] += local_force
      else:
        continue
    return contact_forces

  def compute_jacobian_for_one_leg(self, leg_id: int) -> np.ndarray:
    """Compute the Jacobian for a given leg.

    Args:
      leg_id: Index of the leg for which the jacobian is computed.

    Returns:
      The 3 x N transposed Jacobian matrix. where N is the total DoFs of the
      robot. For a quadruped, the first 6 columns of the matrix corresponds to
      the CoM translation and rotation. The columns corresponds to a leg can be
      extracted with indices [6 + leg_id * 3: 6 + leg_id * 3 + 3]. Note that
      the jacobian is calculated for motors, which takes motor directions into
      consideration.
    """
    com_dof = self._urdf_loader.com_dof
    foot_link_ids = tuple(self._urdf_loader.get_end_effector_id_dict().values())
    return kinematics_utils.compute_jacobian(
        pybullet_client=self.pybullet_client,
        urdf_id=self.robot_id,
        link_id=foot_link_ids[leg_id],
        all_joint_positions=[
            state[0] for state in self._joint_states
        ]) * np.concatenate([np.ones(com_dof), self._motor_directions])

  def map_contact_force_to_joint_torques(
      self, leg_id: int, contact_force: np.ndarray) -> Dict[int, float]:
    """Maps the foot contact force to the leg joint torques.

    Args:
      leg_id: Index of the leg for which the jacobian is computed.
      contact_force: Desired contact force experted by the leg.

    Returns:
      A dict containing the torques for each motor on the leg.
    """
    foot_link_ids = tuple(self._urdf_loader.get_end_effector_id_dict().values())
    jv = self.compute_jacobian_for_one_leg(leg_id)
    all_motor_torques = np.matmul(contact_force, jv)
    motor_torques = {}
    motors_per_leg = self.num_motors // len(foot_link_ids)
    com_dof = self._urdf_loader.com_dof
    for joint_id in range(leg_id * motors_per_leg,
                          (leg_id + 1) * motors_per_leg):
      motor_torques[joint_id] = all_motor_torques[com_dof + joint_id]

    return motor_torques

  @classmethod
  def get_constants(cls):
    raise NotImplementedError("Not yet implemented!")

  @property
  def timestamp(self):
    return self._clock()

  @property
  def action_space(self):
    return self._action_space

  @property
  def action_names(self):
    return self._action_names

  @property
  def base_orientation_quaternion(self):
    """Gets the base orientation as a quaternion.

    The base orientation is always relative to the init_orientation, which
    can be updated by Reset function. This is necessary as many URDF can have an
    internal frame that is not z-up, so if we don't provide an init_orientation
    (through Reset), the loaded robot can have its belly facing the horizontal
    direction.

    Returns:
      The base orientation in quaternion.
    """
    return self._base_orientation_quat

  @property
  def base_orientation_quaternion_default_frame(self):
    """Gets the base orientation in the robot's default frame.

    This is the base orientation in whatever frame the robot specifies. For
    simulated robot this is the URDF's internal frame. For real robot this can
    be based on the rpy reading determined by the IMU.

    Returns:
      The base orientation in quaternion in a robot default frame.
    """
    _, base_orientation_quat = (
        self._pybullet_client.getBasePositionAndOrientation(
            self._urdf_loader.robot_id))
    return base_orientation_quat

  @property
  def sensors(self):
    return self._sensors

  @property
  def base_roll_pitch_yaw(self):
    return self._base_roll_pitch_yaw

  @property
  def base_roll_pitch_yaw_rate(self):
    return self._base_roll_pitch_yaw_rate

  @property
  def base_position(self):
    return self._base_position

  @property
  def base_velocity(self):
    return self._base_velocity

  @property
  def is_safe(self):
    return True

  @property
  def num_motors(self):
    return self._num_motors

  @property
  def motor_model(self):
    return self._motor_model

  @property
  def motor_limits(self) -> robot_config.MotorLimits:
    return self._motor_limits

  @property
  def motor_angles(self):
    return self._motor_angles

  @property
  def motor_velocities(self):
    return self._motor_velocities

  @property
  def motor_torques(self):
    return self._motor_torques

  @property
  def pybullet_client(self):
    return self._pybullet_client

  @property
  def urdf_loader(self):
    return self._urdf_loader

  @property
  def robot_id(self):
    return self._urdf_loader.robot_id

  @property
  def initital_orientation_inverse_quaternion(self):
    return self._init_orientation_inv_quat

  @property
  def base_acceleration_accelerometer(self):
    """Get the base acceleration measured by an accelerometer.

    The acceleration is measured in the local frame of a free-falling robot,
    which is consistent with the robot's IMU measurements. Here the
    gravitational acceleration is first added to the acceleration in the world
    frame, which is then converted to the local frame of the robot.

    """
    return np.array(self._last_base_acceleration_accelerometer)

  @property
  def base_acceleration(self):
    """Get the base acceleration in the world frame."""
    return np.array(self._last_base_acceleration_world)



from pybullet_utils.logger import Logger
logger = Logger()
logger.configure_output_file("e:/mylog.txt")
for i in range(10):
  logger.log_tabular("Iteration", 1)
Logger.print2("hello world")

logger.print_tabular()
logger.dump_tabular()



import gym
import pybullet_envs
import time

#env = gym.make('HumanoidDeepMimicBackflipBulletEnv-v1')
env = gym.make('HumanoidDeepMimicWalkBulletEnv-v1')
env.render(mode='human')
env.reset()
print("------------------------------------")
print("env=",env)
print(dir(env))
print(dir(env.env))
dt = 1./240.
logId = env.env._internal_env._pybullet_client.startStateLogging(env.env._internal_env._pybullet_client.STATE_LOGGING_PROFILE_TIMINGS, "perf.json")
for i in range (100):
  env.env._internal_env._pybullet_client.submitProfileTiming("loop")
  
  env.reset()
  env.env._internal_env._pybullet_client.submitProfileTiming()
env.env._internal_env._pybullet_client.stopStateLogging(logId)

action = env.env.action_space.sample()
while (1):
  time.sleep(dt)
  #keys = env.env._internal_env._pybullet_client.getKeyboardEvents()
  #if keys:
  #  env.reset()
  #action=[0]*36
  action = env.env.action_space.sample()
  state, reward, done, info = env.step(action)
  #env.render(mode='rgb_array')
  if done:
    env.reset()
    #action = env.env.action_space.sample()
  #print("reward=",reward)
    


# Lint as: python3
r"""An example where Laikago walks forward using mpc controller.

This script illustrates an example of MPCLocomotionWrapper class for
controlling a simulated Laikago robot to walk on a flat ground. The wrapped
environment takes as input the target local foothold locations and the desired
base pose. An MPC-based controller is executed internally to compute the
required forces to achieve the desired foothold position and base movement.


blaze run -c opt \
//robotics/reinforcement_learning/minitaur/envs_v2/examples\
:laikago_mpc_wrapper_example
"""
import os
import tempfile

from absl import app
from absl import flags
import gin
import numpy as np
import pybullet_data as pd

from pybullet_envs.minitaur.envs_v2 import env_loader

FLAGS = flags.FLAGS
flags.DEFINE_string("video_file", None, "The filename for saving the videos.")

CONFIG_FILE_SIM = (pd.getDataPath()+"/configs/laikago_mpc_two_camera_random_stepstone.gin")

NUM_STEPS = 100
ENABLE_RENDERING = True  # Will be disabled for tests
ENV_RANDOM_SEED = 100
DEFAULT_TARGET_FOOTHOLD = (0.05, 0.0, -0.01)
DEFAULT_BASE_VELOCITY = (0.0, 0.0)
DEFAULT_TWIST_SPEED = 0.0
DEFAULT_BODY_HEIGHT = 0.45
DEFAULT_ROLL_PITCH = (0.0, 0.0)
DEFAULT_SWING_HEIGHT = 0.07


def _build_env():
  """Builds the environment for the Laikago robot.

  Returns:
    The OpenAI gym environment.
  """
  gin.parse_config_file(CONFIG_FILE_SIM)
  gin.bind_parameter("SimulationParameters.enable_rendering", ENABLE_RENDERING)
  env = env_loader.load()
  env.seed(ENV_RANDOM_SEED)

  return env


def _run_example():
  """An example that Laikago moves with a constant speed and predicts foothold.

  Returns:
    env: the environment after the simulation
  """

  env = _build_env()

  env.reset()
  if FLAGS.video_file is not None:
    pybullet = env.pybullet_client
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)
    log_id = pybullet.startStateLogging(pybullet.STATE_LOGGING_VIDEO_MP4,
                                        FLAGS.video_file)

  try:
    max_step = NUM_STEPS
    for _ in range(max_step):
      target_foothold = np.array(DEFAULT_TARGET_FOOTHOLD)
      action = np.concatenate([
          target_foothold, target_foothold, target_foothold, target_foothold,
          [
              DEFAULT_SWING_HEIGHT, DEFAULT_SWING_HEIGHT, DEFAULT_SWING_HEIGHT,
              DEFAULT_SWING_HEIGHT
          ], DEFAULT_BASE_VELOCITY, [DEFAULT_TWIST_SPEED],
          [DEFAULT_BODY_HEIGHT], DEFAULT_ROLL_PITCH
      ])
      _ = env.step(action)
  finally:
    if FLAGS.video_dir is not None:
      pybullet.stopStateLogging(log_id)


def main(argv):
  del argv
  _run_example()


if __name__ == "__main__":
  app.run(main)



# Code adapted from https://github.com/DLR-RM/rl-baselines3-zoo
# it requires stable-baselines3 to be installed
# Colab Notebook: https://colab.research.google.com/github/Stable-Baselines-Team/rl-colab-notebooks/blob/sb3/pybullet.ipynb
# You can run it using: python -m pybullet_envs.stable_baselines.train --algo td3 --env HalfCheetahBulletEnv-v0
# Author: Antonin RAFFIN
# MIT License
import argparse

import pybullet_envs  # register pybullet envs

import gym
import numpy as np
from stable_baselines3 import SAC, TD3
from stable_baselines3.common.noise import NormalActionNoise
from stable_baselines3.common.callbacks import EvalCallback, CheckpointCallback
from stable_baselines3.common.monitor import Monitor


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Train an RL agent using Stable Baselines3")
    parser.add_argument(
        "--algo",
        help="RL Algorithm (Soft Actor-Critic by default)",
        default="sac",
        type=str,
        required=False,
        choices=["sac", "td3"],
    )
    parser.add_argument(
        "--env", type=str, default="HalfCheetahBulletEnv-v0", help="environment ID"
    )
    parser.add_argument(
        "-n",
        "--n-timesteps",
        help="Number of training timesteps",
        default=int(1e6),
        type=int,
    )
    parser.add_argument(
        "--save-freq",
        help="Save the model every n steps (if negative, no checkpoint)",
        default=-1,
        type=int,
    )
    args = parser.parse_args()

    env_id = args.env
    n_timesteps = args.n_timesteps
    save_path = f"{args.algo}_{env_id}"

    # Instantiate and wrap the environment
    env = gym.make(env_id)

    # Create the evaluation environment and callbacks
    eval_env = Monitor(gym.make(env_id))

    callbacks = [EvalCallback(eval_env, best_model_save_path=save_path)]

    # Save a checkpoint every n steps
    if args.save_freq > 0:
        callbacks.append(
            CheckpointCallback(
                save_freq=args.save_freq, save_path=save_path, name_prefix="rl_model"
            )
        )

    algo = {
        "sac": SAC,
        "td3": TD3,
    }[args.algo]

    n_actions = env.action_space.shape[0]

    # Tuned hyperparameters from https://github.com/DLR-RM/rl-baselines3-zoo
    hyperparams = {
        "sac": dict(
            batch_size=256,
            gamma=0.98,
            policy_kwargs=dict(net_arch=[256, 256]),
            learning_starts=10000,
            buffer_size=int(3e5),
            tau=0.01,
        ),
        "td3": dict(
            batch_size=100,
            policy_kwargs=dict(net_arch=[400, 300]),
            learning_rate=1e-3,
            learning_starts=10000,
            buffer_size=int(1e6),
            train_freq=1,
            gradient_steps=1,
            action_noise=NormalActionNoise(
                mean=np.zeros(n_actions), sigma=0.1 * np.ones(n_actions)
            ),
        ),
    }[args.algo]

    model = algo("MlpPolicy", env, verbose=1, **hyperparams)
    try:
        model.learn(n_timesteps, callback=callbacks)
    except KeyboardInterrupt:
        pass

    print(f"Saving to {save_path}.zip")
    model.save(save_path)



r"""An example to run of the minitaur gym environment with sine gaits.
"""

import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import numpy as np
from pybullet_envs.bullet import minitaur_gym_env
import argparse
from pybullet_envs.bullet import minitaur_env_randomizer


def ResetPoseExample():
  """An example that the minitaur stands still using the reset pose."""
  steps = 1000
  randomizer = (minitaur_env_randomizer.MinitaurEnvRandomizer())
  environment = minitaur_gym_env.MinitaurBulletEnv(render=True,
                                                   leg_model_enabled=False,
                                                   motor_velocity_limit=np.inf,
                                                   pd_control_enabled=True,
                                                   accurate_motor_model_enabled=True,
                                                   motor_overheat_protection=True,
                                                   env_randomizer=randomizer,
                                                   hard_reset=False)
  action = [math.pi / 2] * 8
  for _ in range(steps):
    _, _, done, _ = environment.step(action)
    if done:
      break
  environment.reset()


def MotorOverheatExample():
  """An example of minitaur motor overheat protection is triggered.

  The minitaur is leaning forward and the motors are getting obove threshold
  torques. The overheat protection will be triggered in ~1 sec.
  """

  environment = minitaur_gym_env.MinitaurBulletEnv(render=True,
                                                   leg_model_enabled=False,
                                                   motor_velocity_limit=np.inf,
                                                   motor_overheat_protection=True,
                                                   accurate_motor_model_enabled=True,
                                                   motor_kp=1.20,
                                                   motor_kd=0.00,
                                                   on_rack=False)

  action = [.0] * 8
  for i in range(8):
    action[i] = .0 - 0.1 * (-1 if i % 2 == 0 else 1) * (-1 if i < 4 else 1)

  steps = 500
  actions_and_observations = []
  for step_counter in range(steps):
    # Matches the internal timestep.
    time_step = 0.01
    t = step_counter * time_step
    current_row = [t]
    current_row.extend(action)

    observation, _, _, _ = environment.step(action)
    current_row.extend(observation.tolist())
    actions_and_observations.append(current_row)
  environment.reset()


def SineStandExample():
  """An example of minitaur standing and squatting on the floor.

  To validate the accurate motor model we command the robot and sit and stand up
  periodically in both simulation and experiment. We compare the measured motor
  trajectories, torques and gains.
  """
  environment = minitaur_gym_env.MinitaurBulletEnv(render=True,
                                                   leg_model_enabled=False,
                                                   motor_velocity_limit=np.inf,
                                                   motor_overheat_protection=True,
                                                   accurate_motor_model_enabled=True,
                                                   motor_kp=1.20,
                                                   motor_kd=0.02,
                                                   on_rack=False)
  steps = 1000
  amplitude = 0.5
  speed = 3

  actions_and_observations = []

  for step_counter in range(steps):
    # Matches the internal timestep.
    time_step = 0.01
    t = step_counter * time_step
    current_row = [t]

    action = [math.sin(speed * t) * amplitude + math.pi / 2] * 8
    current_row.extend(action)

    observation, _, _, _ = environment.step(action)
    current_row.extend(observation.tolist())
    actions_and_observations.append(current_row)

  environment.reset()


def SinePolicyExample():
  """An example of minitaur walking with a sine gait."""
  randomizer = (minitaur_env_randomizer.MinitaurEnvRandomizer())
  environment = minitaur_gym_env.MinitaurBulletEnv(render=True,
                                                   motor_velocity_limit=np.inf,
                                                   pd_control_enabled=True,
                                                   hard_reset=False,
                                                   env_randomizer=randomizer,
                                                   on_rack=False)
  sum_reward = 0
  steps = 20000
  amplitude_1_bound = 0.1
  amplitude_2_bound = 0.1
  speed = 1

  for step_counter in range(steps):
    time_step = 0.01
    t = step_counter * time_step

    amplitude1 = amplitude_1_bound
    amplitude2 = amplitude_2_bound
    steering_amplitude = 0
    if t < 10:
      steering_amplitude = 0.1
    elif t < 20:
      steering_amplitude = -0.1
    else:
      steering_amplitude = 0

    # Applying asymmetrical sine gaits to different legs can steer the minitaur.
    a1 = math.sin(t * speed) * (amplitude1 + steering_amplitude)
    a2 = math.sin(t * speed + math.pi) * (amplitude1 - steering_amplitude)
    a3 = math.sin(t * speed) * amplitude2
    a4 = math.sin(t * speed + math.pi) * amplitude2
    action = [a1, a2, a2, a1, a3, a4, a4, a3]
    _, reward, done, _ = environment.step(action)
    sum_reward += reward
    if done:
      break
  environment.reset()


def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--env',
                      help='environment ID (0==sine, 1==stand, 2=reset, 3=overheat)',
                      type=int,
                      default=0)
  args = parser.parse_args()
  print("--env=" + str(args.env))

  if (args.env == 0):
    SinePolicyExample()
  if (args.env == 1):
    SineStandExample()
  if (args.env == 2):
    ResetPoseExample()
  if (args.env == 3):
    MotorOverheatExample()


if __name__ == '__main__':
  main()



import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import gym
import time
from gym import spaces
from gym.utils import seeding
import numpy as np
import pybullet
from . import racecar
import random
from pybullet_utils import bullet_client as bc
import pybullet_data
from pkg_resources import parse_version

RENDER_HEIGHT = 720
RENDER_WIDTH = 960


class RacecarGymEnv(gym.Env):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self,
               urdfRoot=pybullet_data.getDataPath(),
               actionRepeat=50,
               isEnableSelfCollision=True,
               isDiscrete=False,
               renders=False):
    print("init")
    self._timeStep = 0.01
    self._urdfRoot = urdfRoot
    self._actionRepeat = actionRepeat
    self._isEnableSelfCollision = isEnableSelfCollision
    self._observation = []
    self._ballUniqueId = -1
    self._envStepCounter = 0
    self._renders = renders
    self._isDiscrete = isDiscrete
    if self._renders:
      self._p = bc.BulletClient(connection_mode=pybullet.GUI)
    else:
      self._p = bc.BulletClient()

    self.seed()
    #self.reset()
    observationDim = 2  #len(self.getExtendedObservation())
    #print("observationDim")
    #print(observationDim)
    # observation_high = np.array([np.finfo(np.float32).max] * observationDim)
    observation_high = np.ones(observationDim) * 1000  #np.inf
    if (isDiscrete):
      self.action_space = spaces.Discrete(9)
    else:
      action_dim = 2
      self._action_bound = 1
      action_high = np.array([self._action_bound] * action_dim)
      self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self.observation_space = spaces.Box(-observation_high, observation_high, dtype=np.float32)
    self.viewer = None

  def reset(self):
    self._p.resetSimulation()
    #p.setPhysicsEngineParameter(numSolverIterations=300)
    self._p.setTimeStep(self._timeStep)
    #self._p.loadURDF(os.path.join(self._urdfRoot,"plane.urdf"))
    stadiumobjects = self._p.loadSDF(os.path.join(self._urdfRoot, "stadium.sdf"))
    #move the stadium objects slightly above 0
    #for i in stadiumobjects:
    #	pos,orn = self._p.getBasePositionAndOrientation(i)
    #	newpos = [pos[0],pos[1],pos[2]-0.1]
    #	self._p.resetBasePositionAndOrientation(i,newpos,orn)

    dist = 5 + 2. * random.random()
    ang = 2. * 3.1415925438 * random.random()

    ballx = dist * math.sin(ang)
    bally = dist * math.cos(ang)
    ballz = 1

    self._ballUniqueId = self._p.loadURDF(os.path.join(self._urdfRoot, "sphere2.urdf"),
                                          [ballx, bally, ballz])
    self._p.setGravity(0, 0, -10)
    self._racecar = racecar.Racecar(self._p, urdfRootPath=self._urdfRoot, timeStep=self._timeStep)
    self._envStepCounter = 0
    for i in range(100):
      self._p.stepSimulation()
    self._observation = self.getExtendedObservation()
    return np.array(self._observation)

  def __del__(self):
    self._p = 0

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def getExtendedObservation(self):
    self._observation = []  #self._racecar.getObservation()
    carpos, carorn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
    ballpos, ballorn = self._p.getBasePositionAndOrientation(self._ballUniqueId)
    invCarPos, invCarOrn = self._p.invertTransform(carpos, carorn)
    ballPosInCar, ballOrnInCar = self._p.multiplyTransforms(invCarPos, invCarOrn, ballpos, ballorn)

    self._observation.extend([ballPosInCar[0], ballPosInCar[1]])
    return self._observation

  def step(self, action):
    if (self._renders):
      basePos, orn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
      #self._p.resetDebugVisualizerCamera(1, 30, -40, basePos)

    if (self._isDiscrete):
      fwd = [-1, -1, -1, 0, 0, 0, 1, 1, 1]
      steerings = [-0.6, 0, 0.6, -0.6, 0, 0.6, -0.6, 0, 0.6]
      forward = fwd[action]
      steer = steerings[action]
      realaction = [forward, steer]
    else:
      realaction = action

    self._racecar.applyAction(realaction)
    for i in range(self._actionRepeat):
      self._p.stepSimulation()
      if self._renders:
        time.sleep(self._timeStep)
      self._observation = self.getExtendedObservation()

      if self._termination():
        break
      self._envStepCounter += 1
    reward = self._reward()
    done = self._termination()
    #print("len=%r" % len(self._observation))

    return np.array(self._observation), reward, done, {}

  def render(self, mode='human', close=False):
    if mode != "rgb_array":
      return np.array([])
    base_pos, orn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
    view_matrix = self._p.computeViewMatrixFromYawPitchRoll(cameraTargetPosition=base_pos,
                                                            distance=self._cam_dist,
                                                            yaw=self._cam_yaw,
                                                            pitch=self._cam_pitch,
                                                            roll=0,
                                                            upAxisIndex=2)
    proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
                                                     aspect=float(RENDER_WIDTH) / RENDER_HEIGHT,
                                                     nearVal=0.1,
                                                     farVal=100.0)
    (_, _, px, _, _) = self._p.getCameraImage(width=RENDER_WIDTH,
                                              height=RENDER_HEIGHT,
                                              viewMatrix=view_matrix,
                                              projectionMatrix=proj_matrix,
                                              renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)
    rgb_array = np.array(px)
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def _termination(self):
    return self._envStepCounter > 1000

  def _reward(self):
    closestPoints = self._p.getClosestPoints(self._racecar.racecarUniqueId, self._ballUniqueId,
                                             10000)

    numPt = len(closestPoints)
    reward = -1000
    #print(numPt)
    if (numPt > 0):
      #print("reward:")
      reward = -closestPoints[0][8]
      #print(reward)
    return reward

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



from pybullet_envs.bullet.kukaGymEnv import KukaGymEnv
import random
import os
from gym import spaces
import time
import pybullet as p
from . import kuka
import numpy as np
import pybullet_data
import pdb
import distutils.dir_util
import glob
from pkg_resources import parse_version
import gym


class KukaDiverseObjectEnv(KukaGymEnv):
  """Class for Kuka environment with diverse objects.

  In each episode some objects are chosen from a set of 1000 diverse objects.
  These 1000 objects are split 90/10 into a train and test set.
  """

  def __init__(self,
               urdfRoot=pybullet_data.getDataPath(),
               actionRepeat=80,
               isEnableSelfCollision=True,
               renders=False,
               isDiscrete=False,
               maxSteps=8,
               dv=0.06,
               removeHeightHack=False,
               blockRandom=0.3,
               cameraRandom=0,
               width=48,
               height=48,
               numObjects=5,
               isTest=False):
    """Initializes the KukaDiverseObjectEnv.

    Args:
      urdfRoot: The diretory from which to load environment URDF's.
      actionRepeat: The number of simulation steps to apply for each action.
      isEnableSelfCollision: If true, enable self-collision.
      renders: If true, render the bullet GUI.
      isDiscrete: If true, the action space is discrete. If False, the
        action space is continuous.
      maxSteps: The maximum number of actions per episode.
      dv: The velocity along each dimension for each action.
      removeHeightHack: If false, there is a "height hack" where the gripper
        automatically moves down for each action. If true, the environment is
        harder and the policy chooses the height displacement.
      blockRandom: A float between 0 and 1 indicated block randomness. 0 is
        deterministic.
      cameraRandom: A float between 0 and 1 indicating camera placement
        randomness. 0 is deterministic.
      width: The image width.
      height: The observation image height.
      numObjects: The number of objects in the bin.
      isTest: If true, use the test set of objects. If false, use the train
        set of objects.
    """

    self._isDiscrete = isDiscrete
    self._timeStep = 1. / 240.
    self._urdfRoot = urdfRoot
    self._actionRepeat = actionRepeat
    self._isEnableSelfCollision = isEnableSelfCollision
    self._observation = []
    self._envStepCounter = 0
    self._renders = renders
    self._maxSteps = maxSteps
    self.terminated = 0
    self._cam_dist = 1.3
    self._cam_yaw = 180
    self._cam_pitch = -40
    self._dv = dv
    self._p = p
    self._removeHeightHack = removeHeightHack
    self._blockRandom = blockRandom
    self._cameraRandom = cameraRandom
    self._width = width
    self._height = height
    self._numObjects = numObjects
    self._isTest = isTest

    if self._renders:
      self.cid = p.connect(p.SHARED_MEMORY)
      if (self.cid < 0):
        self.cid = p.connect(p.GUI)
      p.resetDebugVisualizerCamera(1.3, 180, -41, [0.52, -0.2, -0.33])
    else:
      self.cid = p.connect(p.DIRECT)
    self.seed()

    if (self._isDiscrete):
      if self._removeHeightHack:
        self.action_space = spaces.Discrete(9)
      else:
        self.action_space = spaces.Discrete(7)
    else:
      self.action_space = spaces.Box(low=-1, high=1, shape=(3,))  # dx, dy, da
      if self._removeHeightHack:
        self.action_space = spaces.Box(low=-1, high=1, shape=(4,))  # dx, dy, dz, da
    self.observation_space = spaces.Box(low=0, high=255, shape=(self._height,
                                                                self._width,
                                                                3))
    self.viewer = None

  def reset(self):
    """Environment reset called at the beginning of an episode.
    """
    # Set the camera settings.
    look = [0.23, 0.2, 0.54]
    distance = 1.
    pitch = -56 + self._cameraRandom * np.random.uniform(-3, 3)
    yaw = 245 + self._cameraRandom * np.random.uniform(-3, 3)
    roll = 0
    self._view_matrix = p.computeViewMatrixFromYawPitchRoll(look, distance, yaw, pitch, roll, 2)
    fov = 20. + self._cameraRandom * np.random.uniform(-2, 2)
    aspect = self._width / self._height
    near = 0.01
    far = 10
    self._proj_matrix = p.computeProjectionMatrixFOV(fov, aspect, near, far)

    self._attempted_grasp = False
    self._env_step = 0
    self.terminated = 0

    p.resetSimulation()
    p.setPhysicsEngineParameter(numSolverIterations=150)
    p.setTimeStep(self._timeStep)
    p.loadURDF(os.path.join(self._urdfRoot, "plane.urdf"), [0, 0, -1])

    p.loadURDF(os.path.join(self._urdfRoot, "table/table.urdf"), 0.5000000, 0.00000, -.820000,
               0.000000, 0.000000, 0.0, 1.0)

    p.setGravity(0, 0, -10)
    self._kuka = kuka.Kuka(urdfRootPath=self._urdfRoot, timeStep=self._timeStep)
    self._envStepCounter = 0
    p.stepSimulation()

    # Choose the objects in the bin.
    urdfList = self._get_random_object(self._numObjects, self._isTest)
    self._objectUids = self._randomly_place_objects(urdfList)
    self._observation = self._get_observation()
    return np.array(self._observation)

  def _randomly_place_objects(self, urdfList):
    """Randomly places the objects in the bin.

    Args:
      urdfList: The list of urdf files to place in the bin.

    Returns:
      The list of object unique ID's.
    """

    # Randomize positions of each object urdf.
    objectUids = []
    for urdf_name in urdfList:
      xpos = 0.4 + self._blockRandom * random.random()
      ypos = self._blockRandom * (random.random() - .5)
      angle = np.pi / 2 + self._blockRandom * np.pi * random.random()
      orn = p.getQuaternionFromEuler([0, 0, angle])
      urdf_path = os.path.join(self._urdfRoot, urdf_name)
      uid = p.loadURDF(urdf_path, [xpos, ypos, .15], [orn[0], orn[1], orn[2], orn[3]])
      objectUids.append(uid)
      # Let each object fall to the tray individual, to prevent object
      # intersection.
      for _ in range(500):
        p.stepSimulation()
    return objectUids

  def _get_observation(self):
    """Return the observation as an image.
    """
    img_arr = p.getCameraImage(width=self._width,
                               height=self._height,
                               viewMatrix=self._view_matrix,
                               projectionMatrix=self._proj_matrix)
    rgb = img_arr[2]
    np_img_arr = np.reshape(rgb, (self._height, self._width, 4))
    return np_img_arr[:, :, :3]

  def step(self, action):
    """Environment step.

    Args:
      action: 5-vector parameterizing XYZ offset, vertical angle offset
      (radians), and grasp angle (radians).
    Returns:
      observation: Next observation.
      reward: Float of the per-step reward as a result of taking the action.
      done: Bool of whether or not the episode has ended.
      debug: Dictionary of extra information provided by environment.
    """
    dv = self._dv  # velocity per physics step.
    if self._isDiscrete:
      # Static type assertion for integers.
      assert isinstance(action, int)
      if self._removeHeightHack:
        dx = [0, -dv, dv, 0, 0, 0, 0, 0, 0][action]
        dy = [0, 0, 0, -dv, dv, 0, 0, 0, 0][action]
        dz = [0, 0, 0, 0, 0, -dv, dv, 0, 0][action]
        da = [0, 0, 0, 0, 0, 0, 0, -0.25, 0.25][action]
      else:
        dx = [0, -dv, dv, 0, 0, 0, 0][action]
        dy = [0, 0, 0, -dv, dv, 0, 0][action]
        dz = -dv
        da = [0, 0, 0, 0, 0, -0.25, 0.25][action]
    else:
      dx = dv * action[0]
      dy = dv * action[1]
      if self._removeHeightHack:
        dz = dv * action[2]
        da = 0.25 * action[3]
      else:
        dz = -dv
        da = 0.25 * action[2]

    return self._step_continuous([dx, dy, dz, da, 0.3])

  def _step_continuous(self, action):
    """Applies a continuous velocity-control action.

    Args:
      action: 5-vector parameterizing XYZ offset, vertical angle offset
      (radians), and grasp angle (radians).
    Returns:
      observation: Next observation.
      reward: Float of the per-step reward as a result of taking the action.
      done: Bool of whether or not the episode has ended.
      debug: Dictionary of extra information provided by environment.
    """
    # Perform commanded action.
    self._env_step += 1
    self._kuka.applyAction(action)
    for _ in range(self._actionRepeat):
      p.stepSimulation()
      if self._renders:
        time.sleep(self._timeStep)
      if self._termination():
        break

    # If we are close to the bin, attempt grasp.
    state = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaEndEffectorIndex)
    end_effector_pos = state[0]
    if end_effector_pos[2] <= 0.1:
      finger_angle = 0.3
      for _ in range(500):
        grasp_action = [0, 0, 0, 0, finger_angle]
        self._kuka.applyAction(grasp_action)
        p.stepSimulation()
        #if self._renders:
        #  time.sleep(self._timeStep)
        finger_angle -= 0.3 / 100.
        if finger_angle < 0:
          finger_angle = 0
      for _ in range(500):
        grasp_action = [0, 0, 0.001, 0, finger_angle]
        self._kuka.applyAction(grasp_action)
        p.stepSimulation()
        if self._renders:
          time.sleep(self._timeStep)
        finger_angle -= 0.3 / 100.
        if finger_angle < 0:
          finger_angle = 0
      self._attempted_grasp = True
    observation = self._get_observation()
    done = self._termination()
    reward = self._reward()

    debug = {'grasp_success': self._graspSuccess}
    return observation, reward, done, debug

  def _reward(self):
    """Calculates the reward for the episode.

    The reward is 1 if one of the objects is above height .2 at the end of the
    episode.
    """
    reward = 0
    self._graspSuccess = 0
    for uid in self._objectUids:
      pos, _ = p.getBasePositionAndOrientation(uid)
      # If any block is above height, provide reward.
      if pos[2] > 0.2:
        self._graspSuccess += 1
        reward = 1
        break
    return reward

  def _termination(self):
    """Terminates the episode if we have tried to grasp or if we are above
    maxSteps steps.
    """
    return self._attempted_grasp or self._env_step >= self._maxSteps

  def _get_random_object(self, num_objects, test):
    """Randomly choose an object urdf from the random_urdfs directory.

    Args:
      num_objects:
        Number of graspable objects.

    Returns:
      A list of urdf filenames.
    """
    if test:
      urdf_pattern = os.path.join(self._urdfRoot, 'random_urdfs/*0/*.urdf')
    else:
      urdf_pattern = os.path.join(self._urdfRoot, 'random_urdfs/*[1-9]/*.urdf')
    found_object_directories = glob.glob(urdf_pattern)
    total_num_objects = len(found_object_directories)
    selected_objects = np.random.choice(np.arange(total_num_objects), num_objects)
    selected_objects_filenames = []
    for object_index in selected_objects:
      selected_objects_filenames += [found_object_directories[object_index]]
    return selected_objects_filenames

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _reset = reset
    _step = step



# Lint as: python3
r"""Tests for locomotion_gym_env.


"""

import math
import random

import gin
import mock
import numpy as np
import unittest
from absl.testing import parameterized

from pybullet_envs.minitaur.envs_v2 import locomotion_gym_env
from pybullet_envs.minitaur.envs_v2.evaluation import metric as metric_lib
from pybullet_envs.minitaur.envs_v2.scenes import scene_base
from pybullet_envs.minitaur.envs_v2.scenes import simple_scene
from pybullet_envs.minitaur.envs_v2.tasks import task_interface
from pybullet_envs.minitaur.envs_v2.utilities import env_utils
from pybullet_envs.minitaur.robots import autonomous_object
from pybullet_envs.minitaur.robots import minitaur_v2
import pybullet_data as pd

import unittest


_POSITION_GAIN = 1.0
_VELOCITY_GAIN = 0.015
_CONTROL_LATENCY = 0.015
_CONFIG_FILE = (pd.getDataPath()+'/configs/minitaur_gym_env.gin')
_CONFIG_FILE_NEW_ROBOT = (pd.getDataPath()+'/configs_v2/base/laikago_reactive.gin')


class TestTask(task_interface.Task):
  """A step counter task for test purpose."""

  def __init__(self):
    self._counter = 0

  def reset(self, env):
    del env
    self._counter = 0

  def reward(self, env):
    del env  # TODO(b/154635313): resolve the API mismatch
    self._counter += 1
    return self._counter

  def update(self, env):
    del env  # TODO(b/154635313): resolve the API mismatch

  def done(self, env):
    del env  # TODO(b/154635313): resolve the API mismatch
    return False


class LocomotionGymEnvTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    gin.clear_config()

  def test_env_from_gin(self):
    # TODO(sehoonha) rename locomotion_gym_*test.gin to locomotion_gym_*.gin
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv()
    self.assertIsInstance(env.robot, minitaur_v2.Minitaur)
    # The robot will stand on the ground.
    self.assertAlmostEqual(env.robot.base_position[2], 0.25, 1)

  def test_reset_gym(self):
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv(task=None)

    desired_init_motor_angle = math.pi / 2
    action_dim = len(env.action_space.high)
    observations = env.reset(initial_motor_angles=[desired_init_motor_angle] *
                             action_dim)
    observations = env_utils.flatten_observations(observations)
    self.assertEqual(observations.size, 12)
    self.assertAlmostEqual(observations[0], 0, 1)
    self.assertAlmostEqual(observations[4], desired_init_motor_angle, 0)

  def test_step_gym(self):
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv(task=TestTask())

    desired_motor_angle = math.pi / 3
    steps = 1000
    action_dim = len(env.action_space.high)
    for _ in range(steps):
      observations, reward, done, _ = env.step([desired_motor_angle] *
                                               action_dim)
      observations = env_utils.flatten_observations(observations)

    self.assertFalse(done)
    self.assertEqual(reward, steps)
    self.assertEqual(observations.size, 12)
    self.assertAlmostEqual(observations[0], 0, 1)
    self.assertAlmostEqual(observations[4], desired_motor_angle, 1)
    np.testing.assert_allclose(env._last_action,
                               [desired_motor_angle] * action_dim, 2e-1)

  def test_scene(self):
    gin.parse_config_file(_CONFIG_FILE)
    data_root = 'urdf/'
    env = locomotion_gym_env.LocomotionGymEnv(
        task=TestTask(), scene=simple_scene.SimpleScene(data_root=data_root))
    desired_motor_angle = math.pi / 3
    steps = 2
    action_dim = len(env.action_space.high)
    for _ in range(steps):
      _, reward, _, _ = env.step([desired_motor_angle] * action_dim)
    self.assertEqual(reward, steps)

  def test_except_on_invalid_config(self):
    gin.parse_config_file(_CONFIG_FILE)
    gin.bind_parameter('SimulationParameters.num_action_repeat', 0)
    with self.assertRaises(ValueError):
      locomotion_gym_env.LocomotionGymEnv(task=None)

  def test_no_scene(self):
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv(task=None, scene=None)

    # The robot will free fall.
    self.assertAlmostEqual(env.robot.base_position[2], 0.15, 1)

  def test_seed_draw_with_np(self):
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv(task=None)
    # first draw
    env.seed(42)
    nums1 = []
    for _ in range(3):
      nums1.append(env.np_random.randint(2**31 - 1))
    # second draw
    env.seed(42)
    nums2 = []
    for _ in range(3):
      nums2.append(env.np_random.randint(2**31 - 1))
    self.assertListEqual(nums1, nums2)

  def test_get_observations(self):
    gin.parse_config_file(_CONFIG_FILE)
    env = locomotion_gym_env.LocomotionGymEnv(task=None)
    desired_init_motor_angle = math.pi / 2
    action_dim = len(env.action_space.high)
    observations = env.reset(initial_motor_angles=[desired_init_motor_angle] *
                             action_dim)
    self.assertEqual(len(observations), 2)
    self.assertEqual(len(observations['IMU']), 4)
    self.assertAlmostEqual(observations['IMU'][0], 0, 2)
    self.assertEqual(len(observations['MotorAngle']), 8)
    self.assertAlmostEqual(observations['MotorAngle'][0], desired_init_motor_angle,0)

  

  

  def test_step_with_dynamic_objects(self):
    gin.parse_config_file(_CONFIG_FILE_NEW_ROBOT)

    gin.parse_config([
        'AutonomousObject.urdf_file = "urdf/mug.urdf"',
        'SceneBase.dynamic_objects = [@AutonomousObject(), @AutonomousObject()]',
        'LocomotionGymEnv.scene = @SceneBase()',
    ])
    env = locomotion_gym_env.LocomotionGymEnv()

    self.assertEqual(len(env.scene.dynamic_objects), 2)
    for obj in env.scene.dynamic_objects:
      self.assertIsInstance(obj, autonomous_object.AutonomousObject)

    # Replace dynamic objects with mocks for step tests.
    mock_objects = [
        mock.create_autospec(autonomous_object.AutonomousObject),
        mock.create_autospec(autonomous_object.AutonomousObject)
    ]
    env.scene._type_to_objects_dict[
        scene_base.ObjectType.DYNAMIC_OBJECT] = mock_objects
    env.step(env.robot.action_space.sample())

    expected_update_calls = [
        mock.call(i * env.sim_time_step, mock.ANY)
        for i in range(env.num_action_repeat)
    ]
    expected_apply_action_calls = [
        mock.call(autonomous_object.AUTONOMOUS_ACTION)
        for i in range(env.num_action_repeat)
    ]
    expected_receive_observation_calls = [
        mock.call() for i in range(env.num_action_repeat)
    ]

    for mock_obj in mock_objects:
      mock_obj.pre_control_step.assert_called_once_with(
          autonomous_object.AUTONOMOUS_ACTION)
      self.assertEqual(mock_obj.update.call_args_list, expected_update_calls)
      self.assertEqual(mock_obj.apply_action.call_args_list,
                       expected_apply_action_calls)
      self.assertEqual(mock_obj.receive_observation.call_args_list,
                       expected_receive_observation_calls)
      mock_obj.post_control_step.assert_called_once_with()

 
  def test_env_metrics(self):
    gin.parse_config_file(_CONFIG_FILE_NEW_ROBOT)
    env = locomotion_gym_env.LocomotionGymEnv()
    metric_logger = env.metric_logger
    test_metric_1 = metric_logger.create_scalar_metric(
        'test_metric_1',
        scope=metric_lib.MetricScope.DEBUG,
        single_ep_aggregator=np.mean)
    test_metric_1.report(22)

    test_metric_2 = metric_logger.create_scalar_metric(
        'test_metric_2',
        scope=metric_lib.MetricScope.PERFORMANCE,
        single_ep_aggregator=np.max)
    test_metric_2.report(15)
    test_metric_2.report(16)

    all_metrics = metric_logger.get_episode_metrics()

    self.assertDictEqual(all_metrics, {
        'DEBUG/test_metric_1': 22,
        'PERFORMANCE/test_metric_2': 16
    })

    env.reset()

    all_metrics = metric_logger.get_episode_metrics()
    self.assertDictEqual(all_metrics, {})


if __name__ == '__main__':
   unittest.main()



import pybullet as p
import time

p.connect(p.GRAPHICS_SERVER_TCP)
import pybullet_data as pd
p.setAdditionalSearchPath(pd.getDataPath())

p.loadURDF("plane.urdf")
p.loadURDF("r2d2.urdf", [0,0,3])
p.setGravity(0,0,-10)
while p.isConnected():
  p.stepSimulation()

